"""
Defines a function which takes a path to an audio file and turns it into text.
"""

from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

from datetime import datetime
import os
import contextlib
import tempfile
import ffmpeg
import subprocess
import openai
from openai import OpenAI

from ..utils.logs import setup_logging
from ..utils.logs import logger
setup_logging()

client = OpenAI()

def convert_mime_type_to_format(mime_type: str) -> str:
    if mime_type == "audio/x-wav" or mime_type == "audio/wav":
        return "wav"
    if mime_type == "audio/webm":
        return "webm"

    return mime_type

@contextlib.contextmanager
def export_audio_to_wav_ffmpeg(audio: bytearray, mime_type: str) -> str:
    temp_dir = tempfile.gettempdir()

    # Create a temporary file with the appropriate extension
    input_ext = convert_mime_type_to_format(mime_type)
    input_path = os.path.join(temp_dir, f"input_{datetime.now().strftime('%Y%m%d%H%M%S%f')}.{input_ext}")
    with open(input_path, 'wb') as f:
        f.write(audio)

    # Check if the input file exists
    assert os.path.exists(input_path), f"Input file does not exist: {input_path}"

    # Export to wav
    output_path = os.path.join(temp_dir, f"output_{datetime.now().strftime('%Y%m%d%H%M%S%f')}.wav")
    ffmpeg.input(input_path).output(output_path, acodec='pcm_s16le', ac=1, ar='16k').run()

    try:
        yield output_path
    finally:
        os.remove(input_path)
        os.remove(output_path)

def run_command(command):
    result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    return result.stdout, result.stderr

def get_transcription_file(wav_file_path: str):
    local_path = os.path.join(os.path.dirname(__file__), 'local_service')
    whisper_rust_path = os.path.join(os.path.dirname(__file__), 'whisper-rust', 'target', 'release')
    model_name = os.getenv('WHISPER_MODEL_NAME')
    if not model_name:
        raise EnvironmentError("WHISPER_MODEL_NAME environment variable is not set.")

    output, error = run_command([
        os.path.join(whisper_rust_path, 'whisper-rust'),
        '--model-path', os.path.join(local_path, model_name),
        '--file-path', wav_file_path
    ])

    print("Transcription result:", output)
    return output

def get_transcription_bytes(audio_bytes: bytearray, mime_type):
    with export_audio_to_wav_ffmpeg(audio_bytes, mime_type) as wav_file_path:
        return get_transcription_file(wav_file_path)

def stt_bytes(audio_bytes: bytearray, mime_type="audio/wav"):
    with export_audio_to_wav_ffmpeg(audio_bytes, mime_type) as wav_file_path:
        return stt_wav(wav_file_path)

def stt_wav(wav_file_path: str):

    if os.getenv('ALL_LOCAL') == 'False':
        audio_file = open(wav_file_path, "rb")
        try:
            transcript = client.audio.transcriptions.create(
                model="whisper-1", 
                file=audio_file,
                response_format="text"
            )
        except openai.BadRequestError as e:
            logger.info(f"openai.BadRequestError: {e}")
            return None

        logger.info(f"Transcription result: {transcript}")
        return transcript
    else:
        temp_dir = tempfile.gettempdir()
        output_path = os.path.join(temp_dir, f"output_stt_{datetime.now().strftime('%Y%m%d%H%M%S%f')}.wav")
        ffmpeg.input(wav_file_path).output(output_path, acodec='pcm_s16le', ac=1, ar='16k').run()
        try:
            transcript = get_transcription_file(output_path)
            print("Transcription result:", transcript)
        finally:
            os.remove(output_path)
        return transcript

def stt(input_data, mime_type="audio/wav"):
    if isinstance(input_data, str):
        return stt_wav(input_data)
    elif isinstance(input_data, bytearray):
        return stt_bytes(input_data, mime_type)
    else:
        raise ValueError("Input data should be either a path to a wav file (str) or audio bytes (bytearray)")


from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import os
import subprocess
from pathlib import Path

### LLM SETUP

# Define the path to a llamafile
llamafile_path = Path(__file__).parent / 'model.llamafile'

# Check if the new llamafile exists, if not download it
if not os.path.exists(llamafile_path):
    subprocess.run(
        [
            "wget",
            "-O",
            llamafile_path,
            "https://huggingface.co/jartine/phi-2-llamafile/resolve/main/phi-2.Q4_K_M.llamafile",
        ],
        check=True,
    )
    # Make the new llamafile executable
    subprocess.run(["chmod", "+x", llamafile_path], check=True)

# Run the new llamafile
subprocess.run([str(llamafile_path)], check=True)


import threading
from datetime import datetime
import json
import subprocess


def _add_message_to_queue(message):
    # Define the message data and convert it to JSON
    message_json = json.dumps({
        "role": "computer",
        "type": "console",
        "format": "output",
        "content": message
    })
    subprocess.run(['logger', '{TO_INTERPRETER{' + message_json + '}TO_INTERPRETER}'])


def schedule(dt: datetime, message: str) -> None:
    """"Schedules a reminder at a specific time. At the specified time, the message will be added to the queue."""
    # Calculate the delay in seconds
    delay = (dt - datetime.now()).total_seconds()

    # Create a timer
    timer = threading.Timer(delay, _add_message_to_queue, args=[message])

    # Start the timer
    timer.start()


"""
This is just for the Python package — we need a Python entrypoint.
Just starts `start.sh` with all the same command line arguments. Aliased to 01.
"""

import os
import subprocess
import sys

def main():

    # Get command line arguments
    args = sys.argv[1:]

    # Get the directory of the current script
    dir_path = os.path.dirname(os.path.realpath(__file__))

    # Prepare the command
    command = [os.path.join(dir_path, 'start.sh')] + args

    # Start start.sh with the command line arguments
    try:
        subprocess.run(command, check=True)
    except KeyboardInterrupt:
        print("Exiting...")
    


from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import asyncio
import subprocess
import platform

from .logs import setup_logging
from .logs import logger
setup_logging()

def get_kernel_messages():
    """
    Is this the way to do this?
    """
    current_platform = platform.system()
    
    if current_platform == "Darwin":
        process = subprocess.Popen(['syslog'], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
        output, _ = process.communicate()
        return output.decode('utf-8')
    elif current_platform == "Linux":
        with open('/var/log/dmesg', 'r') as file:
            return file.read()
    else:
        logger.info("Unsupported platform.")

def custom_filter(message):
    # Check for {TO_INTERPRETER{ message here }TO_INTERPRETER} pattern
    if '{TO_INTERPRETER{' in message and '}TO_INTERPRETER}' in message:
        start = message.find('{TO_INTERPRETER{') + len('{TO_INTERPRETER{')
        end = message.find('}TO_INTERPRETER}', start)
        return message[start:end]
    # Check for USB mention
    # elif 'USB' in message:
    #     return message
    # # Check for network related keywords
    # elif any(keyword in message for keyword in ['network', 'IP', 'internet', 'LAN', 'WAN', 'router', 'switch']) and "networkStatusForFlags" not in message:
        
    #     return message
    else:
        return None
    
last_messages = ""

def check_filtered_kernel():
    messages = get_kernel_messages()
    messages.replace(last_messages, "")
    messages = messages.split("\n")
    
    filtered_messages = []
    for message in messages:
        if custom_filter(message):
            filtered_messages.append(message)
    
    return "\n".join(filtered_messages)

async def put_kernel_messages_into_queue(queue):
    while True:
        text = check_filtered_kernel()
        if text:
            if isinstance(queue, asyncio.Queue):
                await queue.put({"role": "computer", "type": "console", "start": True})
                await queue.put({"role": "computer", "type": "console", "format": "output", "content": text})
                await queue.put({"role": "computer", "type": "console", "end": True})
            else:
                queue.put({"role": "computer", "type": "console", "start": True})
                queue.put({"role": "computer", "type": "console", "format": "output", "content": text})
                queue.put({"role": "computer", "type": "console", "end": True})
        
        await asyncio.sleep(5)


from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import os
import glob
import json
from pathlib import Path
from interpreter import OpenInterpreter


def configure_interpreter(interpreter: OpenInterpreter):
    ### SYSTEM MESSAGE

    # The system message is where most of the 01's behavior is configured.
    # You can put code into the system message {{ in brackets like this }} which will be rendered just before the interpreter starts writing a message.

    system_message = """

    You are an executive assistant AI that helps the user manage their tasks. You can run Python code.

    Store the user's tasks in a Python list called `tasks`.

    ---

    The user's current task is: {{ tasks[0] if tasks else "No current tasks." }}

    {{ 
    if len(tasks) > 1:
    print("The next task is: ", tasks[1])
    }}

    ---

    When the user completes the current task, you should remove it from the list and read the next item by running `tasks = tasks[1:]\ntasks[0]`. Then, tell the user what the next task is.

    When the user tells you about a set of tasks, you should intelligently order tasks, batch similar tasks, and break down large tasks into smaller tasks (for this, you should consult the user and get their permission to break it down). Your goal is to manage the task list as intelligently as possible, to make the user as efficient and non-overwhelmed as possible. They will require a lot of encouragement, support, and kindness. Don't say too much about what's ahead of them— just try to focus them on each step at a time.

    After starting a task, you should check in with the user around the estimated completion time to see if the task is completed. Use the `schedule(datetime, message)` function, which has already been imported.

    To do this, schedule a reminder based on estimated completion time using the function `schedule(datetime_object, "Your message here.")`, WHICH HAS ALREADY BEEN IMPORTED. YOU DON'T NEED TO IMPORT THE `schedule` FUNCTION. IT IS AVALIABLE. You'll recieve the message at `datetime_object`.

    You guide the user through the list one task at a time, convincing them to move forward, giving a pep talk if need be. Your job is essentially to answer "what should I (the user) be doing right now?" for every moment of the day.

    Remember: You can run Python code. Be very concise. Ensure that you actually run code every time! THIS IS IMPORTANT. You NEED to write code. **Help the user by being very concise in your answers.** Do not break down tasks excessively, just into simple, few minute steps. Don't assume the user lives their life in a certain way— pick very general tasks if you're breaking a task down.

    Use the following functions (assume they're imported) to complete your goals whenever possible:
    {{
import sys

original_stdout = sys.stdout
sys.stdout = open(os.devnull, 'w')
original_stderr = sys.stderr
sys.stderr = open(os.devnull, 'w')

from interpreter import interpreter
from pathlib import Path

query = "all functions"
skills_path = Path().resolve() / 'skills'
paths_in_skills = [str(path) for path in skills_path.glob('**/*.py')]
skills = interpreter.computer.docs.search(query, paths=paths_in_skills)
lowercase_skills = [skill[0].lower() + skill[1:] for skill in skills]
output = "\\n".join(lowercase_skills)

sys.stdout = original_stdout
sys.stderr = original_stderr

print(output)
    }}

    """.strip()

    interpreter.custom_instructions = system_message

    ### LLM SETTINGS

    # Local settings
    # interpreter.llm.model = "local"
    # interpreter.llm.api_base = "https://localhost:8080/v1" # Llamafile default
    # interpreter.llm.max_tokens = 1000
    # interpreter.llm.context_window = 3000

    # Hosted settings
    interpreter.llm.api_key = os.getenv('OPENAI_API_KEY')
    interpreter.llm.model = "gpt-4"
    interpreter.auto_run = True
    interpreter.force_task_completion = False


    ### MISC SETTINGS

    interpreter.offline = True
    interpreter.id = 206 # Used to identify itself to other interpreters. This should be changed programatically so it's unique.

    ### RESET conversations/user.json

    script_dir = os.path.dirname(os.path.abspath(__file__))
    user_json_path = os.path.join(script_dir, 'conversations', 'user.json')
    with open(user_json_path, 'w') as file:
        json.dump([], file)

    ### SKILLS
    try:
        interpreter.computer.skills.skills_dir = Path(__file__).parent / 'skills'
        interpreter.computer.skills.import_skills()
    except:
        print("Temporarily skipping skills (OI 0.2.1, which is unreleased) so we can push to `pip`.")
        pass

    return interpreter


from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import os
import asyncio
import threading
import os
import pyaudio
from starlette.websockets import WebSocket
from queue import Queue
from pynput import keyboard
import json
import traceback
import websockets
import queue
import pydub
import ast
from pydub import AudioSegment
from pydub.playback import play
import io
import time
import wave
import tempfile
from datetime import datetime
import cv2
import base64
from interpreter import interpreter # Just for code execution. Maybe we should let people do from interpreter.computer import run?
# In the future, I guess kernel watching code should be elsewhere? Somewhere server / client agnostic?
from ..server.utils.kernel import put_kernel_messages_into_queue
from ..server.utils.get_system_info import get_system_info
from ..server.stt.stt import stt_wav

from ..server.utils.logs import setup_logging
from ..server.utils.logs import logger
setup_logging()


from ..utils.accumulator import Accumulator

accumulator = Accumulator()

# Configuration for Audio Recording
CHUNK = 1024  # Record in chunks of 1024 samples
FORMAT = pyaudio.paInt16  # 16 bits per sample
CHANNELS = 1  # Mono
RATE = 44100  # Sample rate
RECORDING = False  # Flag to control recording state
SPACEBAR_PRESSED = False  # Flag to track spacebar press state

# Camera configuration
CAMERA_ENABLED = bool(os.getenv('CAMERA_ENABLED', False))
CAMERA_DEVICE_INDEX = int(os.getenv('CAMERA_DEVICE_INDEX', 0))
CAMERA_WARMUP_SECONDS = float(os.getenv('CAMERA_WARMUP_SECONDS', 0))

# Specify OS
current_platform = get_system_info()

# Initialize PyAudio
p = pyaudio.PyAudio()

send_queue = queue.Queue()

class Device:
    def __init__(self):
        self.pressed_keys = set()
        self.captured_images = []
        self.audiosegments = []

    def fetch_image_from_camera(self, camera_index=CAMERA_DEVICE_INDEX):
        """Captures an image from the specified camera device and saves it to a temporary file. Adds the image to the captured_images list."""
        image_path = None

        cap = cv2.VideoCapture(camera_index)
        ret, frame = cap.read()  # Capture a single frame to initialize the camera

        if CAMERA_WARMUP_SECONDS > 0:
            # Allow camera to warm up, then snap a picture again
            # This is a workaround for some cameras that don't return a properly exposed
            # picture immediately when they are first turned on
            time.sleep(CAMERA_WARMUP_SECONDS)
            ret, frame = cap.read()

        if ret:
            temp_dir = tempfile.gettempdir()
            image_path = os.path.join(temp_dir, f"01_photo_{datetime.now().strftime('%Y%m%d%H%M%S%f')}.png")
            self.captured_images.append(image_path)
            cv2.imwrite(image_path, frame)
            logger.info(f"Camera image captured to {image_path}")
            logger.info(f"You now have {len(self.captured_images)} images which will be sent along with your next audio message.")
        else:
            logger.error(f"Error: Couldn't capture an image from camera ({camera_index})")

        cap.release()

        return image_path
    

    def encode_image_to_base64(self, image_path):
        """Encodes an image file to a base64 string."""
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def add_image_to_send_queue(self, image_path):
        """Encodes an image and adds an LMC message to the send queue with the image data."""
        base64_image = self.encode_image_to_base64(image_path)
        image_message = {
            "role": "user",
            "type": "image",
            "format": "base64.png",
            "content": base64_image
        }
        send_queue.put(image_message)
        # Delete the image file from the file system after sending it
        os.remove(image_path)

    def queue_all_captured_images(self):
        """Queues all captured images to be sent."""
        for image_path in self.captured_images:
            self.add_image_to_send_queue(image_path)
        self.captured_images.clear()  # Clear the list after sending

        
    async def play_audiosegments(self):
        """Plays them sequentially."""
        while True:
            try:
                for audio in self.audiosegments:
                    play(audio)
                    self.audiosegments.remove(audio)
                await asyncio.sleep(0.1)
            except asyncio.exceptions.CancelledError:
                # This happens once at the start?
                pass
            except:
                logger.info(traceback.format_exc())


    def record_audio(self):
        
        if os.getenv('STT_RUNNER') == "server":
            # STT will happen on the server. we're sending audio.
            send_queue.put({"role": "user", "type": "audio", "format": "bytes.wav", "start": True})
        elif os.getenv('STT_RUNNER') == "client":
            # STT will happen here, on the client. we're sending text.
            send_queue.put({"role": "user", "type": "message", "start": True})
        else:
            raise Exception("STT_RUNNER must be set to either 'client' or 'server'.")

        """Record audio from the microphone and add it to the queue."""
        stream = p.open(format=FORMAT, channels=CHANNELS, rate=RATE, input=True, frames_per_buffer=CHUNK)
        logger.info("Recording started...")
        global RECORDING

        # Create a temporary WAV file to store the audio data
        temp_dir = tempfile.gettempdir()
        wav_path = os.path.join(temp_dir, f"audio_{datetime.now().strftime('%Y%m%d%H%M%S%f')}.wav")
        wav_file = wave.open(wav_path, 'wb')
        wav_file.setnchannels(CHANNELS)
        wav_file.setsampwidth(p.get_sample_size(FORMAT))
        wav_file.setframerate(RATE)

        while RECORDING:
            data = stream.read(CHUNK, exception_on_overflow=False)
            wav_file.writeframes(data)

        wav_file.close()
        stream.stop_stream()
        stream.close()
        logger.info("Recording stopped.")

        duration = wav_file.getnframes() / RATE
        if duration < 0.3:
            # Just pressed it. Send stop message
            if os.getenv('STT_RUNNER') == "client":
                send_queue.put({"role": "user", "type": "message", "content": "stop"})
                send_queue.put({"role": "user", "type": "message", "end": True})
            else:
                send_queue.put({"role": "user", "type": "audio", "format": "bytes.wav", "content": ""})
                send_queue.put({"role": "user", "type": "audio", "format": "bytes.wav", "end": True})
        else:
            self.queue_all_captured_images()

            if os.getenv('STT_RUNNER') == "client":
                # Run stt then send text
                text = stt_wav(wav_path)
                send_queue.put({"role": "user", "type": "message", "content": text})
                send_queue.put({"role": "user", "type": "message", "end": True})
            else:
                # Stream audio
                with open(wav_path, 'rb') as audio_file:
                    byte_data = audio_file.read(CHUNK)
                    while byte_data:
                        send_queue.put(byte_data)
                        byte_data = audio_file.read(CHUNK)
                send_queue.put({"role": "user", "type": "audio", "format": "bytes.wav", "end": True})

        if os.path.exists(wav_path):
            os.remove(wav_path)

    def toggle_recording(self, state):
        """Toggle the recording state."""
        global RECORDING, SPACEBAR_PRESSED
        if state and not SPACEBAR_PRESSED:
            SPACEBAR_PRESSED = True
            if not RECORDING:
                RECORDING = True
                threading.Thread(target=self.record_audio).start()
        elif not state and SPACEBAR_PRESSED:
            SPACEBAR_PRESSED = False
            RECORDING = False

    def on_press(self, key):
        """Detect spacebar press and Ctrl+C combination."""
        self.pressed_keys.add(key)  # Add the pressed key to the set

        if keyboard.Key.space in self.pressed_keys:
            self.toggle_recording(True)
        elif {keyboard.Key.ctrl, keyboard.KeyCode.from_char('c')} <= self.pressed_keys:
            logger.info("Ctrl+C pressed. Exiting...")
            os._exit(0)

    def on_release(self, key):
        """Detect spacebar release and 'c' key press for camera, and handle key release."""
        self.pressed_keys.discard(key)  # Remove the released key from the key press tracking set

        if key == keyboard.Key.space:
            self.toggle_recording(False)
        elif CAMERA_ENABLED and key == keyboard.KeyCode.from_char('c'):
            self.fetch_image_from_camera()

    
    async def message_sender(self, websocket):
        while True:
            message = await asyncio.get_event_loop().run_in_executor(None, send_queue.get)
            if isinstance(message, bytes):
                await websocket.send(message)
            else:
                await websocket.send(json.dumps(message))
            send_queue.task_done()
            await asyncio.sleep(0.01)

    async def websocket_communication(self, WS_URL):
        while True:
            try:
                async with websockets.connect(WS_URL) as websocket:
                    if CAMERA_ENABLED:
                        logger.info("Press the spacebar to start/stop recording. Press 'c' to capture an image from the camera. Press CTRL-C to exit.")
                    else:
                        logger.info("Press the spacebar to start/stop recording. Press CTRL-C to exit.")
                        
                    asyncio.create_task(self.message_sender(websocket))

                    while True:
                        await asyncio.sleep(0.01)
                        chunk = await websocket.recv()

                        logger.debug(f"Got this message from the server: {type(chunk)} {chunk}")

                        if type(chunk) == str:
                            chunk = json.loads(chunk)

                        message = accumulator.accumulate(chunk)
                        if message == None:
                            # Will be None until we have a full message ready
                            continue

                        # At this point, we have our message

                        if message["type"] == "audio" and message["format"].startswith("bytes"):

                            # Convert bytes to audio file
                            # Format will be bytes.wav or bytes.opus
                            audio_bytes = io.BytesIO(message["content"])
                            audio = AudioSegment.from_file(audio_bytes, codec=message["format"].split(".")[1])

                            self.audiosegments.append(audio)

                        # Run the code if that's the client's job
                        if os.getenv('CODE_RUNNER') == "client":
                            if message["type"] == "code" and "end" in message:
                                language = message["format"]
                                code = message["content"]
                                result = interpreter.computer.run(language, code)
                                send_queue.put(result)
    

            except:
                logger.debug(traceback.format_exc())
                logger.info(f"Connecting to `{WS_URL}`...")
                await asyncio.sleep(2)

    async def start_async(self):
            # Configuration for WebSocket
            WS_URL = os.getenv('SERVER_URL')
            if not WS_URL:
                raise ValueError("The environment variable SERVER_URL is not set. Please set it to proceed.")

            # Start the WebSocket communication
            asyncio.create_task(self.websocket_communication(WS_URL))

            # Start watching the kernel if it's your job to do that
            if os.getenv('CODE_RUNNER') == "client":
                asyncio.create_task(put_kernel_messages_into_queue(send_queue))

            asyncio.create_task(self.play_audiosegments())
            
            # If Raspberry Pi, add the button listener, otherwise use the spacebar
            if current_platform.startswith("raspberry-pi"):
                logger.info("Raspberry Pi detected, using button on GPIO pin 15")
                # Use GPIO pin 15
                pindef = ["gpiochip4", "15"] # gpiofind PIN15
                print("PINDEF", pindef)

                # HACK: needs passwordless sudo
                process = await asyncio.create_subprocess_exec("sudo", "gpiomon", "-brf", *pindef, stdout=asyncio.subprocess.PIPE)
                while True:
                    line = await process.stdout.readline()
                    if line:
                        line = line.decode().strip()
                        if "FALLING" in line:
                            self.toggle_recording(False)
                        elif "RISING" in line:
                            self.toggle_recording(True)
                    else:
                        break
            else:
                # Keyboard listener for spacebar press/release
                listener = keyboard.Listener(on_press=self.on_press, on_release=self.on_release)
                listener.start()

    def start(self):
        if os.getenv('TEACH_MODE') != "True":
            asyncio.run(self.start_async())
            p.terminate()


from datetime import datetime
from .utils.logs import setup_logging, logger
import tkinter as tk
import tkinter.simpledialog
from interpreter import interpreter
from tkinter import messagebox
from ..utils.accumulator import Accumulator
import time
import os
import textwrap

setup_logging()
accumulator = Accumulator()
class Skill:
    def __init__(self, name: str):
        self.skill_name = name
        self.steps = []
        self.code = ""

def to_camel_case(text):
    words = text.split()
    camel_case_string = words[0].lower() + ''.join(word.title() for word in words[1:])
    return camel_case_string

def generate_python_code(function_name, code):
    code_string = f'def {to_camel_case(function_name)}():\n'
    code_string += f'    """{function_name}"""\n'
    indented_code = textwrap.indent(code, '    ')
    code_string += indented_code + '\n'
    return code_string

def generate_python_steps(function_name, steps):
    code_string = f'def {to_camel_case(function_name)}():\n'
    code_string += f'    """{function_name}"""\n'
    code_string += f'    print({steps})\n'
    return code_string

def teach():
    root = tk.Tk()
    root.withdraw()

    skill_name = tkinter.simpledialog.askstring("Skill Name", "Please enter the name for the skill:")
    skill = Skill(skill_name)
    while True:
        step = tkinter.simpledialog.askstring("Next Step", "Enter the next step (or 'end' to finish): ")
        logger.info(f"Performing step: {step}")
        if step == "end":
            break

        chunk_code = ""
        interpreter.computer.languages = [l for l in interpreter.computer.languages if l.name.lower() == "python"]
        interpreter.force_task_completion = True
        for chunk in interpreter.chat(step, stream=True, display=False):
            if "format" in chunk and chunk["format"] == "execution":
                content = chunk["content"]
                language = content["format"]
                code = content["content"]
                chunk_code += code
                interpreter.computer.run(code, language)
            time.sleep(0.05)
            accumulator.accumulate(chunk)

        isCorrect = messagebox.askyesno("To Proceed?", "Did I do this step right?")
        if isCorrect:
            skill.steps.append(step)
            skill.code += chunk_code

    # Uncomment this incase you want steps instead of code
    #python_code = generate_python_steps(skill.skill_name, skill.steps)
    
    python_code = generate_python_code(skill.skill_name, skill.code)
    SKILLS_DIR = os.path.dirname(__file__) + "/skills"
    filename = os.path.join(SKILLS_DIR, f"{skill.skill_name.replace(' ', '_')}.py")
    with open(filename, "w") as file:
        file.write(python_code)



"""
Defines a function which takes text and returns a path to an audio file.
"""

from pydub import AudioSegment
from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import tempfile
from openai import OpenAI
import os
import subprocess
import tempfile
from pydub import AudioSegment

client = OpenAI()

chunk_size = 1024

def stream_tts(text):
    """
    A generator that streams tts as LMC messages.
    """
    if os.getenv('ALL_LOCAL') == 'False':
        response = client.audio.speech.create(
            model="tts-1",
            voice="alloy",
            input=text,
            response_format="opus"
        )
        with tempfile.NamedTemporaryFile(suffix=".opus") as temp_file:
            response.stream_to_file(temp_file.name)

            audio_bytes = temp_file.read()
            file_type = "bytes.opus"

    else:
        with tempfile.NamedTemporaryFile(suffix=".wav", delete=False) as temp_file:
            output_file = temp_file.name
            piper_dir = os.path.join(os.path.dirname(__file__), 'local_service', 'piper')
            subprocess.run([
                os.path.join(piper_dir, 'piper'),
                '--model', os.path.join(piper_dir, os.getenv('PIPER_VOICE_NAME')),
                '--output_file', output_file
            ], input=text, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

            audio_bytes = temp_file.read()
            file_type = "bytes.wav"

    # Stream the audio
    yield {"role": "assistant", "type": "audio", "format": file_type, "start": True}
    for i in range(0, len(audio_bytes), chunk_size):
        chunk = audio_bytes[i:i+chunk_size]
        yield chunk
    yield {"role": "assistant", "type": "audio", "format": file_type, "end": True}

def play_audiosegment(audio):
    """
    UNUSED
    the default makes some pops. this fixes that
    """

    # Apply a fade-out (optional but recommended to smooth the end)
    audio = audio.fade_out(500)

    # Add silence at the end
    silence_duration_ms = 500  # Duration of silence in milliseconds
    silence = AudioSegment.silent(duration=silence_duration_ms)
    audio_with_padding = audio + silence

    # Save the modified audio as a WAV file for compatibility with simpleaudio
    audio_with_padding.export("output_audio.wav", format="wav")

    # Load the processed WAV file
    wave_obj = sa.WaveObject.from_wave_file("output_audio.wav")

    # Play the audio
    play_obj = wave_obj.play()

    # Wait for the playback to finish
    play_obj.wait_done()

    # Delete the wav file
    os.remove("output_audio.wav")




from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import ast
import json
import queue
import os
import traceback
import re
from fastapi import FastAPI
from fastapi.responses import PlainTextResponse
from starlette.websockets import WebSocket, WebSocketDisconnect
from .stt.stt import stt_bytes
from .tts.tts import stream_tts
from pathlib import Path
import asyncio
import urllib.parse
from .utils.kernel import put_kernel_messages_into_queue
from .i import configure_interpreter
from interpreter import interpreter
import ngrok
from ..utils.accumulator import Accumulator
from .teach import teach
from .utils.logs import setup_logging
from .utils.logs import logger
setup_logging()

accumulator = Accumulator()

app = FastAPI()

conversation_history_path = Path(__file__).parent / 'conversations' / 'user.json'

# This is so we only say() full sentences
def is_full_sentence(text):
    return text.endswith(('.', '!', '?'))

def split_into_sentences(text):
    return re.split(r'(?<=[.!?])\s+', text)

# Queues
from_computer = queue.Queue() # Just for computer messages from the device. Sync queue because interpreter.run is synchronous
from_user = asyncio.Queue() # Just for user messages from the device.
to_device = asyncio.Queue() # For messages we send.

# Switch code executor to device if that's set

if os.getenv('CODE_RUNNER') == "device":

    # (This should probably just loop through all languages and apply these changes instead)

    class Python:
        # This is the name that will appear to the LLM.
        name = "python"

        def __init__(self):
            self.halt = False

        def run(self, code):
            """Generator that yields a dictionary in LMC Format."""

            # Prepare the data
            message = {"role": "assistant", "type": "code", "format": "python", "content": code}

            # Unless it was just sent to the device, send it wrapped in flags
            if not (interpreter.messages and interpreter.messages[-1] == message):
                to_device.put({"role": "assistant", "type": "code", "format": "python", "start": True})
                to_device.put(message)
                to_device.put({"role": "assistant", "type": "code", "format": "python", "end": True})
            
            # Stream the response
            logger.info("Waiting for the device to respond...")
            while True:
                chunk = from_computer.get()
                logger.info(f"Server received from device: {chunk}")
                if "end" in chunk:
                    break
                yield chunk

        def stop(self):
            self.halt = True

        def terminate(self):
            """Terminates the entire process."""
            # dramatic!! do nothing
            pass

    interpreter.computer.languages = [Python]

# Configure interpreter
interpreter = configure_interpreter(interpreter)

@app.get("/ping")
async def ping():
    return PlainTextResponse("pong")

@app.websocket("/")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    receive_task = asyncio.create_task(receive_messages(websocket))
    send_task = asyncio.create_task(send_messages(websocket))
    try:
        await asyncio.gather(receive_task, send_task)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.info(f"Connection lost. Error: {e}")

async def receive_messages(websocket: WebSocket):
    while True:
        try:
            try:
                data = await websocket.receive()
            except Exception as e:
                print(str(e))
                return
            if 'text' in data:
                try:
                    data = json.loads(data['text'])
                    if data["role"] == "computer":
                        from_computer.put(data) # To be handled by interpreter.computer.run
                    elif data["role"] == "user":
                        await from_user.put(data)
                    else:
                        raise("Unknown role:", data)
                except json.JSONDecodeError:
                    pass  # data is not JSON, leave it as is
            elif 'bytes' in data:
                data = data['bytes']  # binary data
                await from_user.put(data)
        except WebSocketDisconnect as e:
            if e.code == 1000:
                logger.info("Websocket connection closed normally.")
                return
            else:
                raise
        

async def send_messages(websocket: WebSocket):
    while True:
        message = await to_device.get()
        logger.debug(f"Sending to the device: {type(message)} {message}")
        
        try:
            if isinstance(message, dict):
                await websocket.send_json(message)
            elif isinstance(message, bytes):
                await websocket.send_bytes(message)
            else:
                raise TypeError("Message must be a dict or bytes")
        except:
            # Make sure to put the message back in the queue if you failed to send it
            await to_device.put(message)
            raise

async def listener():

    while True:
        while True:
            if not from_user.empty():
                chunk = await from_user.get()
                break
            elif not from_computer.empty():
                chunk = from_computer.get()
                break
            await asyncio.sleep(1)

        

        message = accumulator.accumulate(chunk)
        if message == None:
            # Will be None until we have a full message ready
            continue

        # print(str(message)[:1000])

        # At this point, we have our message

        if message["type"] == "audio" and message["format"].startswith("bytes"):

            if not message["content"]: # If it was nothing / silence
                continue

            # Convert bytes to audio file
            # Format will be bytes.wav or bytes.opus
            mime_type = "audio/" + message["format"].split(".")[1]
            text = stt_bytes(message["content"], mime_type)
            message = {"role": "user", "type": "message", "content": text}

        # At this point, we have only text messages

        # Custom stop message will halt us
        if message["content"].lower().strip(".,! ") == "stop":
            continue

        # Load, append, and save conversation history
        with open(conversation_history_path, 'r') as file:
            messages = json.load(file)
        messages.append(message)
        with open(conversation_history_path, 'w') as file:
            json.dump(messages, file, indent=4)

        accumulated_text = ""
        
        for chunk in interpreter.chat(messages, stream=True, display=False):

            logger.debug("Got chunk:", chunk)

            # Send it to the user
            await to_device.put(chunk)
            # Yield to the event loop, so you actually send it out
            await asyncio.sleep(0.01)
            
            if os.getenv('TTS_RUNNER') == "server":
                # Speak full sentences out loud
                if chunk["role"] == "assistant" and "content" in chunk:
                    accumulated_text += chunk["content"]
                    sentences = split_into_sentences(accumulated_text)
                    
                    # If we're going to speak, say we're going to stop sending text.
                    # This should be fixed probably, we should be able to do both in parallel, or only one.
                    if any(is_full_sentence(sentence) for sentence in sentences):
                        await to_device.put({"role": "assistant", "type": "message", "end": True})
                    
                    if is_full_sentence(sentences[-1]):
                        for sentence in sentences:
                            await stream_tts_to_device(sentence)
                        accumulated_text = ""
                    else:
                        for sentence in sentences[:-1]:
                            await stream_tts_to_device(sentence)
                        accumulated_text = sentences[-1]

                    # If we're going to speak, say we're going to stop sending text.
                    # This should be fixed probably, we should be able to do both in parallel, or only one.
                    if any(is_full_sentence(sentence) for sentence in sentences):
                        await to_device.put({"role": "assistant", "type": "message", "start": True})
                
            # If we have a new message, save our progress and go back to the top
            if not from_user.empty():

                # Check if it's just an end flag. We ignore those.
                temp_message = await from_user.get()
                
                if temp_message == {'role': 'user', 'type': 'message', 'end': True}:
                    # Yup. False alarm.
                    continue
                else:
                    # Whoops! Put that back
                    await from_user.put(temp_message)

                with open(conversation_history_path, 'w') as file:
                    json.dump(interpreter.messages, file, indent=4)

                logger.info("New user message recieved. Breaking.")
                break

            # Also check if there's any new computer messages
            if not from_computer.empty():
                
                with open(conversation_history_path, 'w') as file:
                    json.dump(interpreter.messages, file, indent=4)

                logger.info("New computer message recieved. Breaking.")
                break
        else:
            with open(conversation_history_path, 'w') as file:
                json.dump(interpreter.messages, file, indent=4)  

async def stream_tts_to_device(sentence):
    for chunk in stream_tts(sentence):
        await to_device.put(chunk)
        
async def setup_ngrok(ngrok_auth_token, parsed_url):
    # Set up Ngrok
    logger.info("Setting up Ngrok")
    ngrok_listener = await ngrok.forward(f"{parsed_url.hostname}:{parsed_url.port}", authtoken=ngrok_auth_token)
    ngrok_parsed_url = urllib.parse.urlparse(ngrok_listener.url())

    # Setup SERVER_URL environment variable for device to use
    connection_url = f"wss://{ngrok_parsed_url.hostname}/"
    logger.info(f"Ngrok established at {ngrok_parsed_url.geturl()}")
    logger.info(f"\033[1mSERVER_CONNECTION_URL should be set to \"{connection_url}\"\033[0m")


from uvicorn import Config, Server

# Run the FastAPI app
if __name__ == "__main__":

    async def main():
        if os.getenv('TEACH_MODE') == "True":
            teach()
        else:
            # Start listening
            asyncio.create_task(listener())

            # Start watching the kernel if it's your job to do that
            if os.getenv('CODE_RUNNER') == "server":
                asyncio.create_task(put_kernel_messages_into_queue(from_computer))

            server_url = os.getenv('SERVER_URL')
            if not server_url:
                raise ValueError("The environment variable SERVER_URL is not set. Please set it to proceed.")
            parsed_url = urllib.parse.urlparse(server_url)

            # Set up Ngrok
            ngrok_auth_token = os.getenv('NGROK_AUTHTOKEN')
            if ngrok_auth_token is not None:
                await setup_ngrok(ngrok_auth_token, parsed_url)
                
            logger.info("Starting `server.py`...")

            config = Config(app, host=parsed_url.hostname, port=parsed_url.port, lifespan='on')
            server = Server(config)
            await server.serve()

    asyncio.run(main())


from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import os
import logging

logger: logging.Logger = logging.getLogger("01")
root_logger: logging.Logger = logging.getLogger()


def _basic_config() -> None:
    logging.basicConfig(
        format="%(message)s"
    )


def setup_logging() -> None:
    env = os.environ.get("LOG_LEVEL", "").upper()
    if env == "DEBUG":
        _basic_config()
        logger.setLevel(logging.DEBUG)
        root_logger.setLevel(logging.DEBUG)
    elif env == "INFO":
        _basic_config()
        logger.setLevel(logging.INFO)



class Accumulator:
    def __init__(self):
        self.template = {"role": None, "type": None, "format": None, "content": None} 
        self.message = self.template

    def accumulate(self, chunk):
        #print(str(chunk)[:100])
        if type(chunk) == dict:

            if "format" in chunk and chunk["format"] == "active_line":
                # We don't do anything with these
                return None

            if "start" in chunk:
                self.message = chunk
                self.message.pop("start")
                return None

            if "content" in chunk:

                # Display
                print(chunk['content'], end="", flush=True)

                if any(self.message[key] != chunk[key] for key in self.message if key != "content"):
                    self.message = chunk
                if "content" not in self.message:
                    self.message["content"] = chunk["content"]
                else:
                    if type(chunk["content"]) == dict:
                        # dict concatenation cannot happen, so we see if chunk is a dict
                        self.message["content"]["content"] += chunk["content"]["content"]
                    else:
                        self.message["content"] += chunk["content"]
                return None

            if "end" in chunk:
                # We will proceed
                message = self.message
                self.message = self.template
                return message

        if type(chunk) == bytes:
            if "content" not in self.message or type(self.message["content"]) != bytes:
                self.message["content"] = b""
            self.message["content"] += chunk
            return None

        


import os
import platform


def get_system_info():
    system = platform.system()
    
    if system == "Linux":
        # Attempt to identify specific Linux distribution
        distro = "linux"  # Default to generic 'linux'
        try:
            with open("/etc/os-release") as f:
                os_release_info = f.read().lower()
            if "ubuntu" in os_release_info:
                return "raspberry-pi-ubuntu"
            elif "raspbian" in os_release_info:
                return "raspberry-pi-os"
        except FileNotFoundError:
            pass
        
        # Check for Raspberry Pi hardware
        try:
            with open("/proc/device-tree/model") as f:
                model_info = f.read()
            if "Raspberry Pi" in model_info:
                if distro == "ubuntu":
                    return "raspberry-pi-ubuntu"
                return "raspberry-pi"
        except FileNotFoundError:
            pass
        
        return distro
    elif system == "Darwin":
        return "darwin"
    elif system == "Windows":
        return "windows"
    else:
        return "unknown"
